import os
from glob import glob
from io import BytesIO

from fontTools.ttLib import TTFont

from ttfautohint import ttfautohint

import pytest


DATA = os.path.join(os.path.dirname(__file__), "data")
UNHINTED_TTFS = glob(os.path.join(DATA, "*.ttf"))

GLOBAL_HINTING_TABLES = ["fpgm", "prep", "cvt ", "gasp"]


def autohint_font(ttfont, **options):
    buf = BytesIO()
    ttfont.save(buf)
    data = ttfautohint(in_buffer=buf.getvalue(), **options)
    return TTFont(BytesIO(data))


@pytest.fixture(params=UNHINTED_TTFS, ids=lambda p: os.path.basename(p))
def unhinted(request):
    return TTFont(request.param)


class TestTTFAutohint(object):
    def test_simple(self, unhinted):
        for tag in GLOBAL_HINTING_TABLES:
            assert tag not in unhinted
        assert not unhinted["glyf"]["a"].program
        nameID5 = unhinted["name"].getName(5, 3, 1).toUnicode()
        assert "; ttfautohint" not in nameID5

        hinted = autohint_font(unhinted)

        for tag in GLOBAL_HINTING_TABLES:
            assert tag in hinted
        assert hinted["glyf"]["a"].program

        nameID5 = hinted["name"].getName(5, 3, 1).toUnicode()
        assert "; ttfautohint" in nameID5

    def test_in_and_out_file_paths(self, tmpdir, unhinted):
        in_file = tmpdir / "unhinted.ttf"
        out_file = tmpdir / "hinted.ttf"

        with in_file.open("wb") as f:
            unhinted.save(f)

        ttfautohint(in_file=str(in_file), out_file=str(out_file))

        assert os.path.getsize(str(out_file)) > 0

    def test_no_info(self, unhinted):
        hinted = autohint_font(unhinted, no_info=True)

        nameID5 = hinted["name"].getName(5, 3, 1).toUnicode()
        assert "; ttfautohint" not in nameID5

    def test_detailed_info(self, unhinted):
        hinted = autohint_font(unhinted, detailed_info=True)

        nameID5 = hinted["name"].getName(5, 3, 1).toUnicode()
        assert "; ttfautohint" in nameID5

        assert ' -l 8 -r 50 -G 200 -x 14 -D latn -f none -a qsq -X ""' in nameID5

    def test_family_suffix(self, unhinted):
        suffix = " Hinted"
        hinted = autohint_font(unhinted, family_suffix=suffix)

        nametable = hinted["name"]
        nameID1 = nametable.getName(1, 3, 1).toUnicode()

        assert nameID1.endswith(suffix)



import sys
import os
import tempfile
from collections import OrderedDict
from enum import IntEnum
from ttfautohint._compat import ensure_binary, ensure_text

USER_OPTIONS = dict(
    in_file=None,
    in_buffer=None,
    out_file=None,
    control_file=None,
    control_buffer=None,
    reference_file=None,
    reference_buffer=None,
    reference_index=0,
    hinting_range_min=8,
    hinting_range_max=50,
    hinting_limit=200,
    hint_composites=False,
    adjust_subglyphs=False,
    increase_x_height=14,
    x_height_snapping_exceptions="",
    windows_compatibility=False,
    default_script="latn",
    fallback_script="none",
    fallback_scaling=False,
    symbol=False,
    fallback_stem_width=None,
    ignore_restrictions=False,
    family_suffix=None,
    detailed_info=False,
    no_info=False,
    TTFA_info=False,
    dehint=False,
    epoch=None,
    debug=False,
    verbose=False,
)

StemWidthMode = IntEnum(
    "StemWidthMode",
    [
        "NATURAL",  # -1
        "QUANTIZED",  # 0
        "STRONG",  # 1
    ],
    start=-1,
)

STEM_WIDTH_MODE_OPTIONS = OrderedDict(
    [
        ("gray_stem_width_mode", StemWidthMode.QUANTIZED),
        ("gdi_cleartype_stem_width_mode", StemWidthMode.STRONG),
        ("dw_cleartype_stem_width_mode", StemWidthMode.QUANTIZED),
    ]
)

USER_OPTIONS.update(STEM_WIDTH_MODE_OPTIONS)

# Deprecated; use stem width mode options
STRONG_STEM_WIDTH_OPTIONS = dict(
    gdi_cleartype_strong_stem_width=True,
    gray_strong_stem_width=False,
    dw_cleartype_strong_stem_width=False,
)


def validate_options(kwargs):
    opts = {k: kwargs.pop(k, USER_OPTIONS[k]) for k in USER_OPTIONS}
    if kwargs:
        raise TypeError(
            "unknown keyword argument%s: %s"
            % ("s" if len(kwargs) > 1 else "", ", ".join(repr(k) for k in kwargs))
        )

    if opts["no_info"] and opts["detailed_info"]:
        raise ValueError("no_info and detailed_info are mutually exclusive")

    in_file, in_buffer = opts.pop("in_file"), opts.pop("in_buffer")
    if in_file is None and in_buffer is None:
        raise ValueError("No input file or buffer provided")
    elif in_file is not None and in_buffer is not None:
        raise ValueError("in_file and in_buffer are mutually exclusive")
    if in_file is not None:
        try:
            in_buffer = in_file.read()
        except AttributeError:
            with open(in_file, "rb") as f:
                in_buffer = f.read()
    if not isinstance(in_buffer, bytes):
        raise TypeError(
            "in_buffer type must be bytes, not %s" % type(in_buffer).__name__
        )
    opts["in_buffer"] = in_buffer

    control_file = opts.pop("control_file")
    control_buffer = opts.pop("control_buffer")
    if control_buffer is not None:
        if control_file is not None:
            raise ValueError("control_file and control_buffer are mutually exclusive")
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp.write(ensure_binary(control_buffer, "utf-8"))
        control_file = tmp.name
    if control_file is not None:
        opts["control_file"] = control_file

    reference_file = opts.pop("reference_file")
    reference_buffer = opts.pop("reference_buffer")
    if reference_buffer is not None:
        if reference_file is not None:
            raise ValueError(
                "reference_file and reference_buffer are mutually exclusive"
            )
        if not isinstance(reference_buffer, bytes):
            raise TypeError(
                "reference_buffer type must be bytes, not %s"
                % type(reference_buffer).__name__
            )
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp.write(ensure_binary(reference_buffer, "utf-8"))
        reference_file = tmp.name
    if reference_file is not None:
        opts["reference_file"] = reference_file

    if opts["family_suffix"] is not None:
        opts["family_suffix"] = ensure_text(opts["family_suffix"])

    for mode_option in STEM_WIDTH_MODE_OPTIONS:
        # raises ValueError if integer value is not a valid stem width mode
        opts[mode_option] = StemWidthMode(opts[mode_option])

    return opts


# CLI options that have different names from the corresponding ttfautohint()
# keyword parameters.
SPECIAL_OPT_NAMES = {
    "reference_file": "reference",
    "hint_composites": "composites",
    "TTFA_info": "ttfa-table",
}


def format_kwargs(**options):
    # convert keyword parameters to CLI flags suitable for ttfautohint command line
    result = []
    modes = {}
    for key, value in options.items():
        if key not in USER_OPTIONS:
            continue
        if value is None:
            continue
        if key in STEM_WIDTH_MODE_OPTIONS:
            modes[key] = value
            continue
        opt = f"--{SPECIAL_OPT_NAMES.get(key, key).replace('_', '-').lower()}"
        if isinstance(value, bool):
            if value:
                result.append(opt)
        elif isinstance(value, (int, float, str)):
            if value != USER_OPTIONS[key]:
                result.append(opt)
                result.append(str(value))
        else:
            raise TypeError(f"{key}: {type(value)}")
    if modes:
        result.extend(["--stem-width-mode", format_stem_width_modes(**modes)])
    return result


def strong_stem_width(s):
    if len(s) > 3:
        import argparse

        raise argparse.ArgumentTypeError("string can only contain up to 3 letters")
    valid = {
        "g": "gray_stem_width_mode",
        "G": "gdi_cleartype_stem_width_mode",
        "D": "dw_cleartype_stem_width_mode",
    }
    chars = set(s)
    invalid = chars - set(valid)
    if invalid:
        import argparse

        raise argparse.ArgumentTypeError(
            "invalid value: %s" % ", ".join(repr(v) for v in sorted(invalid))
        )
    result = {}
    for char, opt_name in valid.items():
        is_strong = char in chars
        result[opt_name] = (
            StemWidthMode.STRONG if is_strong else StemWidthMode.QUANTIZED
        )
    return result


def stem_width_mode(s):
    if len(s) != 3:
        import argparse

        raise argparse.ArgumentTypeError(
            "Stem width mode string must consist of exactly three letters"
        )
    modes = {k[0].lower(): v for k, v in StemWidthMode.__members__.items()}
    result = {}
    for i, option in enumerate(STEM_WIDTH_MODE_OPTIONS):
        m = s[i]
        if m not in modes:
            import argparse

            letters = sorted(repr(k) for k in modes)
            raise argparse.ArgumentTypeError(
                "Stem width mode letter for %s must be %s, or %s"
                % (option, ", ".join(letters[:-1]), letters[-1])
            )
        result[option] = modes[m]
    return result


def format_stem_width_modes(
    gray_stem_width_mode,
    gdi_cleartype_stem_width_mode,
    dw_cleartype_stem_width_mode,
):
    return (
        gray_stem_width_mode.name[0].lower()
        + gdi_cleartype_stem_width_mode.name[0].lower()
        + dw_cleartype_stem_width_mode.name[0].lower()
    )


def stdin_or_input_path_type(s):
    # the special argument "-" means sys.stdin
    if s == "-":
        try:
            if sys.stdin.isatty():  # ignore if interactive
                return None
            return open(sys.stdin.fileno(), mode="rb", closefd=False)
        except (AttributeError, IOError):
            # if stdout was redirected (e.g. inside pytest), fileno may raise
            # io.UnsupportedOperation
            return None
    return s


def stdout_or_output_path_type(s):
    # the special argument "-" means sys.stdout
    if s == "-":
        try:
            if sys.stdout.isatty():  # ignore if interactive
                return None
            return open(sys.stdout.fileno(), mode="wb", closefd=False)
        except (AttributeError, IOError):
            # if stdout was redirected (e.g. inside pytest), fileno may raise
            # io.UnsupportedOperation
            return None
    return s


def _windows_cmdline2list(cmdline):
    """Build an argv list from a Microsoft shell style cmdline str.

    The reverse of subprocess.list2cmdline that follows the same MS C runtime rules.
    Borrowed from Jython source code:
    https://github.com/jython/jython/blob/50729e6/Lib/subprocess.py#L668-L722
    """
    whitespace = " \t"
    # count of preceding '\'
    bs_count = 0
    in_quotes = False
    arg = []
    argv = []

    for ch in cmdline:
        if ch in whitespace and not in_quotes:
            if arg:
                # finalize arg and reset
                argv.append("".join(arg))
                arg = []
            bs_count = 0
        elif ch == "\\":
            arg.append(ch)
            bs_count += 1
        elif ch == '"':
            if not bs_count % 2:
                # Even number of '\' followed by a '"'. Place one
                # '\' for every pair and treat '"' as a delimiter
                if bs_count:
                    del arg[-(bs_count / 2) :]
                in_quotes = not in_quotes
            else:
                # Odd number of '\' followed by a '"'. Place one '\'
                # for every pair and treat '"' as an escape sequence
                # by the remaining '\'
                del arg[-(bs_count / 2 + 1) :]
                arg.append(ch)
            bs_count = 0
        else:
            # regular char
            arg.append(ch)
            bs_count = 0

    # A single trailing '"' delimiter yields an empty arg
    if arg or in_quotes:
        argv.append("".join(arg))

    return argv


def _parse_ttfautohint_version_string():
    from ttfautohint import run

    result = run(["--version"], capture_output=True, check=True)

    output = result.stdout
    if not output:
        raise ValueError("Could not parse ttfautohint --version")

    first_line = result.stdout.decode("utf-8").splitlines()[0]
    if not first_line.startswith("ttfautohint "):
        raise ValueError(f"ttfautohint --version has unexpected format: {first_line}")

    return first_line[12:]


def parse_args(args=None, splitfunc=None):
    """Parse command line arguments and return a dictionary of options
    for ttfautohint.ttfautohint function.

    `args` can be either None, a list of strings, or a single string,
    that is split into individual options with `shlex.split`.

    When `args` is None, the console's default sys.argv are used, and any
    SystemExit exceptions raised by argparse are propagated.

    If args is a string list or a string, it is assumed that the function
    was not called from a console script's `main` entry point, but from
    other client code, and thus the SystemExit exceptions are muted and
    a `None` value is returned.
    """
    import argparse
    from ttfautohint import __version__
    from ttfautohint.cli import USAGE, DESCRIPTION, EPILOG
    import warnings

    warnings.warn(
        "`ttfautohint.options.parse_args` is deprecated and will be removed "
        "in a future release. Use `ttfautohint.run` instead.",
        DeprecationWarning,
    )

    version_string = "ttfautohint-py %s (libttfautohint %s)" % (
        __version__,
        _parse_ttfautohint_version_string(),
    )

    if args is None:
        capture_sys_exit = False
    else:
        capture_sys_exit = True
        if isinstance(args, str):
            if splitfunc is None:
                if sys.platform == "win32":
                    splitfunc = _windows_cmdline2list
                else:
                    import shlex

                    splitfunc = shlex.split
            args = splitfunc(args)

    parser = argparse.ArgumentParser(
        prog="ttfautohint",
        usage=USAGE,
        description=DESCRIPTION,
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "in_file",
        nargs="?",
        metavar="IN-FILE",
        default="-",
        type=stdin_or_input_path_type,
        help="input file (default: standard input)",
    )
    parser.add_argument(
        "out_file",
        nargs="?",
        metavar="OUT-FILE",
        default="-",
        type=stdout_or_output_path_type,
        help="output file (default: standard output)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="print debugging information"
    )

    stem_width_group = parser.add_mutually_exclusive_group(required=False)
    stem_width_group.add_argument(
        "-a",
        "--stem-width-mode",
        type=stem_width_mode,
        metavar="S",
        default=STEM_WIDTH_MODE_OPTIONS,
        help=(
            "select stem width mode for grayscale, GDI ClearType, and DW "
            "ClearType, where S is a string of three letters with possible "
            "values 'n' for natural, 'q' for quantized, and 's' for strong "
            "(default: qsq)"
        ),
    )
    stem_width_group.add_argument(  # deprecated
        "-w",
        "--strong-stem-width",
        type=strong_stem_width,
        metavar="S",
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "-c",
        "--composites",
        dest="hint_composites",
        action="store_true",
        help="hint glyph composites also",
    )
    parser.add_argument("-d", "--dehint", action="store_true", help="remove all hints")
    parser.add_argument(
        "-D",
        "--default-script",
        metavar="SCRIPT",
        default=USER_OPTIONS["default_script"],
        help="set default OpenType script (default: %(default)s)",
    )
    parser.add_argument(
        "-f",
        "--fallback-script",
        metavar="SCRIPT",
        default=USER_OPTIONS["fallback_script"],
        help="set fallback script (default: %(default)s)",
    )
    parser.add_argument(
        "-F",
        "--family-suffix",
        metavar="SUFFIX",
        help="append SUFFIX to the family name string(s) in the `name' table",
    )
    parser.add_argument(
        "-G",
        "--hinting-limit",
        type=int,
        metavar="PPEM",
        default=USER_OPTIONS["hinting_limit"],
        help=(
            "switch off hinting above this PPEM value (default: "
            "%(default)s); value 0 means no limit"
        ),
    )
    parser.add_argument(
        "-H",
        "--fallback-stem-width",
        type=int,
        metavar="UNITS",
        default=USER_OPTIONS["fallback_stem_width"],
        help=("set fallback stem width (default: 50 font units at 2048 UPEM)"),
    )
    parser.add_argument(
        "-i",
        "--ignore-restrictions",
        action="store_true",
        help="override font license restrictions",
    )
    parser.add_argument(
        "-I",
        "--detailed-info",
        action="store_true",
        help=(
            "add detailed ttfautohint info to the version string(s) in "
            "the `name' table"
        ),
    )
    parser.add_argument(
        "-l",
        "--hinting-range-min",
        type=int,
        metavar="PPEM",
        default=USER_OPTIONS["hinting_range_min"],
        help="the minimum PPEM value for hint sets (default: %(default)s)",
    )
    parser.add_argument(
        "-m",
        "--control-file",
        metavar="FILE",
        help="get control instructions from FILE",
    )
    parser.add_argument(
        "-n",
        "--no-info",
        action="store_true",
        help=(
            "don't add ttfautohint info to the version string(s) in the " "`name' table"
        ),
    )
    parser.add_argument(
        "-p",
        "--adjust-subglyphs",
        action="store_true",
        help="handle subglyph adjustments in exotic fonts",
    )
    parser.add_argument(
        "-r",
        "--hinting-range-max",
        type=int,
        metavar="PPEM",
        default=USER_OPTIONS["hinting_range_max"],
        help="the maximum PPEM value for hint sets (default: %(default)s)",
    )
    parser.add_argument(
        "-R",
        "--reference",
        dest="reference_file",
        metavar="FILE",
        help="derive blue zones from reference font FILE",
    )
    parser.add_argument(
        "-s", "--symbol", action="store_true", help="input is symbol font"
    )
    parser.add_argument(
        "-S",
        "--fallback-scaling",
        action="store_true",
        help="use fallback scaling, not hinting",
    )
    parser.add_argument(
        "-t",
        "--ttfa-table",
        action="store_true",
        dest="TTFA_info",
        help="add TTFA information table",
    )
    parser.add_argument(
        "-T",
        "--ttfa-info",
        dest="show_TTFA_info",
        action="store_true",
        help="display TTFA table in IN-FILE and exit",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="show progress information"
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=version_string,
        help="print version information and exit",
    )
    parser.add_argument(
        "-W",
        "--windows-compatibility",
        action="store_true",
        help=(
            "add blue zones for `usWinAscent' and `usWinDescent' to avoid " "clipping"
        ),
    )
    parser.add_argument(
        "-x",
        "--increase-x-height",
        type=int,
        metavar="PPEM",
        default=USER_OPTIONS["increase_x_height"],
        help=(
            "increase x height for sizes in the range 6<=PPEM<=N; value "
            "0 switches off this feature (default: %(default)s)"
        ),
    )
    parser.add_argument(
        "-X",
        "--x-height-snapping-exceptions",
        metavar="STRING",
        default=USER_OPTIONS["x_height_snapping_exceptions"],
        help=(
            "specify a comma-separated list of x-height snapping exceptions"
            ', for example "-9, 13-17, 19" (default: "%(default)s")'
        ),
    )
    parser.add_argument(
        "-Z",
        "--reference-index",
        type=int,
        metavar="NUMBER",
        default=USER_OPTIONS["reference_index"],
        help="face index of reference font (default: %(default)s)",
    )

    try:
        options = vars(parser.parse_args(args))
    except SystemExit:
        if capture_sys_exit:
            return None
        raise

    # if either input/output are interactive, print help and exit
    if not capture_sys_exit and (
        options["in_file"] is None or options["out_file"] is None
    ):
        parser.print_help()
        parser.exit(1)

    # check SOURCE_DATE_EPOCH environment variable
    source_date_epoch = os.environ.get("SOURCE_DATE_EPOCH")
    if source_date_epoch:
        try:
            options["epoch"] = int(source_date_epoch)
        except ValueError:
            import warnings

            warnings.warn(
                UserWarning("invalid SOURCE_DATE_EPOCH: %r" % source_date_epoch)
            )

    if options.pop("show_TTFA_info"):
        # TODO use fonttools to dump TTFA table?
        raise NotImplementedError()

    stem_width_options = options.pop("stem_width_mode")
    strong_stem_width_options = options.pop("strong_stem_width")
    if strong_stem_width_options:
        import warnings

        warnings.warn(UserWarning("Option '-w' is deprecated! Use option '-a' instead"))
        stem_width_options = strong_stem_width_options
    options.update(stem_width_options)

    return options



import atexit
import io
import os
import stat
import subprocess
import sys
from contextlib import ExitStack
from importlib.resources import as_file, files, is_resource

from ttfautohint._version import __version__
from ttfautohint.errors import TAError
from ttfautohint.options import validate_options, format_kwargs, StemWidthMode


__all__ = ["__version__", "ttfautohint", "TAError", "StemWidthMode", "run"]


# clean up resources on exit
_exit_stack = ExitStack()
atexit.register(_exit_stack.close)

_exe_basename = "ttfautohint"
if sys.platform == "win32":
    _exe_basename += ".exe"
_exe_full_path = None


def _executable_path() -> str:
    global _exe_full_path

    if _exe_full_path is None:
        if is_resource(__name__, _exe_basename):
            _exe_full_path = str(
                _exit_stack.enter_context(
                    as_file(files(__name__).joinpath(_exe_basename))
                )
            )
            # need to chmod +x in case it was extracted from a zip
            if not os.access(_exe_full_path, os.X_OK):
                os.chmod(_exe_full_path, os.stat(_exe_full_path).st_mode | stat.S_IEXEC)
        else:
            import shutil

            _exe_full_path = shutil.which(_exe_basename)
            if _exe_full_path is None:
                raise TAError("ttfautohint executable not found on $PATH")

    return _exe_full_path


def run(args, **kwargs):
    """Run the 'ttfautohint' executable with the list of positional arguments.

    All keyword arguments are forwarded to subprocess.run function.

    The bundled copy of the 'ttfautohint' executable is tried first; if this
    was not included at installation, the version which is on $PATH is used.

    Return:
        subprocess.CompletedProcess object with the following attributes:
        args, returncode, stdout, stderr.
    """
    return subprocess.run([_executable_path()] + list(args), **kwargs)


# TODO: add docstring
def ttfautohint(**kwargs):
    options = validate_options(kwargs)

    in_buffer = options.pop("in_buffer")
    out_file = options.pop("out_file")

    capture_output = True
    stdout = None
    should_close_stdout = False
    if out_file is not None:
        if isinstance(out_file, (str, bytes, os.PathLike)):
            stdout, out_file = open(out_file, "w"), None
            should_close_stdout = True
            capture_output = False
        else:
            try:
                out_file.fileno()
            except io.UnsupportedOperation:
                if not out_file.writable():
                    raise TypeError(f"{out_file} is not writable")
            else:
                stdout, out_file = out_file, None
                capture_output = False

    args = format_kwargs(**options)

    result = run(
        args,
        input=in_buffer,
        capture_output=capture_output,
        stdout=stdout,
    )
    if result.returncode != 0:
        raise TAError(result.returncode, result.stderr)

    output_data = result.stdout

    if output_data and out_file is not None:
        out_file.write(output_data)

    if stdout is not None and should_close_stdout:
        stdout.close()

    return output_data



import sys
from io import StringIO, BytesIO
import argparse
import os
import pytest

from ttfautohint._compat import ensure_binary
from ttfautohint.options import (
    validate_options,
    strong_stem_width,
    stdin_or_input_path_type,
    stdout_or_output_path_type,
    parse_args,
    format_kwargs,
    stem_width_mode,
    StemWidthMode,
    _windows_cmdline2list,
)


class TestValidateOptions(object):
    def test_no_input(self):
        with pytest.raises(ValueError, match="No input file"):
            validate_options({})

    def test_unknown_keyword(self):
        kwargs = dict(foo="bar")
        with pytest.raises(TypeError, match="unknown keyword argument: 'foo'"):
            validate_options(kwargs)

        # 's' for plural
        kwargs = dict(foo="bar", baz=False)
        with pytest.raises(TypeError, match="unknown keyword arguments: 'foo', 'baz'"):
            validate_options(kwargs)

    def test_no_info_or_detailed_info(self, tmpdir):
        msg = "no_info and detailed_info are mutually exclusive"
        kwargs = dict(no_info=True, detailed_info=True)
        with pytest.raises(ValueError, match=msg):
            validate_options(kwargs)

    def test_in_file_or_in_buffer(self, tmpdir):
        msg = "in_file and in_buffer are mutually exclusive"
        in_file = (tmpdir / "file1.ttf").ensure()
        kwargs = dict(in_file=str(in_file), in_buffer=b"\x00\x01\x00\x00")
        with pytest.raises(ValueError, match=msg):
            validate_options(kwargs)

    def test_control_file_or_control_buffer(self, tmpdir):
        msg = "control_file and control_buffer are mutually exclusive"
        control_file = (tmpdir / "ta_ctrl.txt").ensure()
        kwargs = dict(
            in_buffer=b"\0\1\0\0", control_file=control_file, control_buffer=b"abcd"
        )
        with pytest.raises(ValueError, match=msg):
            validate_options(kwargs)

    def test_reference_file_or_reference_buffer(self, tmpdir):
        msg = "reference_file and reference_buffer are mutually exclusive"
        reference_file = (tmpdir / "ref.ttf").ensure()
        kwargs = dict(
            in_buffer=b"\0\1\0\0",
            reference_file=reference_file,
            reference_buffer=b"\x00\x01\x00\x00",
        )
        with pytest.raises(ValueError, match=msg):
            validate_options(kwargs)

    def test_in_file_to_in_buffer(self, tmpdir):
        in_file = tmpdir / "file1.ttf"
        data = b"\0\1\0\0"
        in_file.write_binary(data)

        # 'in_file' is a file-like object
        options = validate_options({"in_file": in_file.open(mode="rb")})
        assert options["in_buffer"] == data
        assert "in_file" not in options

        # 'in_file' is a path string
        options = validate_options({"in_file": str(in_file)})
        assert options["in_buffer"] == data
        assert "in_file" not in options

    def test_in_buffer_is_bytes(self, tmpdir):
        with pytest.raises(TypeError, match="in_buffer type must be bytes"):
            validate_options({"in_buffer": "abcd"})

    def test_control_buffer_to_control_file(self, tmpdir):
        kwargs = {"in_buffer": b"\0", "control_buffer": "abcd"}
        options = validate_options(kwargs)

        assert "control_buffer" not in options
        assert isinstance(options["control_file"], str)
        with open(options["control_file"], "r") as f:
            assert f.read() == "abcd"

    def test_reference_buffer_to_reference_file(self, tmpdir):
        kwargs = {"in_buffer": b"\0", "reference_buffer": b"\0\1\0\0"}
        options = validate_options(kwargs)

        assert "reference_buffer" not in options
        assert isinstance(options["reference_file"], str)
        with open(options["reference_file"], "rb") as f:
            assert f.read() == b"\0\1\0\0"

    def test_reference_buffer_is_bytes(self, tmpdir):
        with pytest.raises(TypeError, match="reference_buffer type must be bytes"):
            validate_options({"in_buffer": b"\0", "reference_buffer": ""})

    def test_epoch(self):
        options = validate_options({"in_buffer": b"\0", "epoch": 0})
        assert isinstance(options["epoch"], int)
        assert options["epoch"] == 0

    def test_family_suffix(self):
        options = validate_options({"in_buffer": b"\0", "family_suffix": b"-TA"})
        assert isinstance(options["family_suffix"], str)
        assert options["family_suffix"] == "-TA"


@pytest.mark.parametrize(
    "options, expected",
    [
        ({}, []),
        (
            {
                "in_file": "/path/to/input_font.ttf",
                "out_file": "/path/to/output_font.ttf",
                "control_file": "/path/to/control_file.txt",
                "reference_file": "/path/to/reference_font.ttf",
                "reference_index": 1,
                "hinting_range_min": 9,
                "hinting_range_max": 51,
                "hinting_limit": 201,
                "hint_composites": True,
                "adjust_subglyphs": True,
                "increase_x_height": 15,
                "x_height_snapping_exceptions": "6,15-18",
                "windows_compatibility": True,
                "default_script": "grek",
                "fallback_script": "latn",
                "fallback_scaling": True,
                "symbol": True,
                "fallback_stem_width": 100,
                "ignore_restrictions": True,
                "family_suffix": "-Hinted",
                "detailed_info": True,
                "no_info": False,
                "TTFA_info": True,
                "dehint": True,
                "epoch": 1513955869,
                "debug": False,
                "verbose": True,
                "gray_stem_width_mode": StemWidthMode.NATURAL,
                "gdi_cleartype_stem_width_mode": StemWidthMode.NATURAL,
                "dw_cleartype_stem_width_mode": StemWidthMode.NATURAL,
            },
            [
                "--in-file",
                "/path/to/input_font.ttf",
                "--out-file",
                "/path/to/output_font.ttf",
                "--control-file",
                "/path/to/control_file.txt",
                "--reference",
                "/path/to/reference_font.ttf",
                "--reference-index",
                "1",
                "--hinting-range-min",
                "9",
                "--hinting-range-max",
                "51",
                "--hinting-limit",
                "201",
                "--composites",
                "--adjust-subglyphs",
                "--increase-x-height",
                "15",
                "--x-height-snapping-exceptions",
                "6,15-18",
                "--windows-compatibility",
                "--default-script",
                "grek",
                "--fallback-script",
                "latn",
                "--fallback-scaling",
                "--symbol",
                "--fallback-stem-width",
                "100",
                "--ignore-restrictions",
                "--family-suffix",
                "-Hinted",
                "--detailed-info",
                "--ttfa-table",
                "--dehint",
                "--epoch",
                "1513955869",
                "--verbose",
                "--stem-width-mode",
                "nnn",
            ],
        ),
        ({"unkown_option": 1}, []),
    ],
    ids=[
        "empty",
        "full-options",
        "unknown-option",
    ],
)
def test_format_kwargs(options, expected):
    assert format_kwargs(**options) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        (
            "",
            {
                "gray_stem_width_mode": StemWidthMode.QUANTIZED,
                "gdi_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
                "dw_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
            },
        ),
        (
            "g",
            {
                "gray_stem_width_mode": StemWidthMode.STRONG,
                "gdi_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
                "dw_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
            },
        ),
        (
            "G",
            {
                "gray_stem_width_mode": StemWidthMode.QUANTIZED,
                "gdi_cleartype_stem_width_mode": StemWidthMode.STRONG,
                "dw_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
            },
        ),
        (
            "D",
            {
                "gray_stem_width_mode": StemWidthMode.QUANTIZED,
                "gdi_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
                "dw_cleartype_stem_width_mode": StemWidthMode.STRONG,
            },
        ),
        (
            "DGg",
            {
                "gray_stem_width_mode": StemWidthMode.STRONG,
                "gdi_cleartype_stem_width_mode": StemWidthMode.STRONG,
                "dw_cleartype_stem_width_mode": StemWidthMode.STRONG,
            },
        ),
    ],
    ids=["empty-string", "only-gray", "only-gdi", "only-dw", "all"],
)
def test_strong_stem_width(string, expected):
    assert strong_stem_width(string) == expected


def test_strong_stem_width_invalid():
    with pytest.raises(
        argparse.ArgumentTypeError, match="string can only contain up to 3 letters"
    ):
        strong_stem_width("GGGG")

    with pytest.raises(argparse.ArgumentTypeError, match="invalid value: 'a'"):
        strong_stem_width("a")


@pytest.mark.parametrize(
    "string, expected",
    [
        (
            "nnn",
            {
                "gray_stem_width_mode": StemWidthMode.NATURAL,
                "gdi_cleartype_stem_width_mode": StemWidthMode.NATURAL,
                "dw_cleartype_stem_width_mode": StemWidthMode.NATURAL,
            },
        ),
        (
            "qqq",
            {
                "gray_stem_width_mode": StemWidthMode.QUANTIZED,
                "gdi_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
                "dw_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
            },
        ),
        (
            "sss",
            {
                "gray_stem_width_mode": StemWidthMode.STRONG,
                "gdi_cleartype_stem_width_mode": StemWidthMode.STRONG,
                "dw_cleartype_stem_width_mode": StemWidthMode.STRONG,
            },
        ),
        (
            "nqs",
            {
                "gray_stem_width_mode": StemWidthMode.NATURAL,
                "gdi_cleartype_stem_width_mode": StemWidthMode.QUANTIZED,
                "dw_cleartype_stem_width_mode": StemWidthMode.STRONG,
            },
        ),
    ],
    ids=["nnn", "qqq", "sss", "nqs"],
)
def test_stem_width_mode(string, expected):
    assert stem_width_mode(string) == expected


def test_stem_width_mode_invalid():
    with pytest.raises(
        argparse.ArgumentTypeError, match="must consist of exactly three letters"
    ):
        stem_width_mode("nnnn")

    with pytest.raises(
        argparse.ArgumentTypeError, match="Stem width mode letter for .* must be"
    ):
        stem_width_mode("zzz")


@pytest.fixture(
    params=[True, False],
    ids=["tty", "pipe"],
)
def isatty(request):
    return request.param


class MockFile(object):
    def __init__(self, f, isatty):
        self._file = f
        self._isatty = isatty

    def isatty(self):
        return self._isatty

    def __getattr__(self, attr):
        return getattr(self._file, attr)


def test_stdin_input_type(monkeypatch, tmpdir, isatty):
    tmp = (tmpdir / "stdin").ensure().open("r")
    monkeypatch.setattr(sys, "stdin", MockFile(tmp, isatty))

    f = stdin_or_input_path_type("-")

    if isatty:
        assert f is None
    else:
        assert hasattr(f, "read")
        assert f.mode == "rb"
        assert f.closed is False


def test_path_input_type(tmpdir):
    tmp = tmpdir / "font.ttf"
    s = str(tmp)
    path = stdin_or_input_path_type(s)
    assert path == s


def test_stdout_output_type(monkeypatch, tmpdir, isatty):
    tmp = (tmpdir / "stdout").open("w")
    monkeypatch.setattr(sys, "stdout", MockFile(tmp, isatty))

    f = stdout_or_output_path_type("-")

    if isatty:
        assert f is None
    else:
        assert hasattr(f, "write")
        assert f.mode == "wb"
        assert f.closed is False


def test_path_output_type(tmpdir):
    tmp = tmpdir / "font.ttf"
    s = str(tmp)
    path = stdout_or_output_path_type(s)
    assert path == s


class TestParseArgs(object):
    argv0 = "python -m ttfautohint"

    def test_unrecognized_arguments(self, monkeypatch, capsys):
        monkeypatch.setattr(argparse._sys, "argv", [self.argv0, "--foo"])

        with pytest.raises(SystemExit) as exc_info:
            parse_args()

        assert str(exc_info.value) == "2"
        assert "unrecognized arguments: --foo" in capsys.readouterr()[1]

        monkeypatch.undo()

        assert parse_args("--bar") is None
        assert "unrecognized arguments: --bar" in capsys.readouterr()[1]

        assert parse_args(["--baz"]) is None
        assert "unrecognized arguments: --baz" in capsys.readouterr()[1]

    def test_no_in_file(self, monkeypatch, capsys):
        monkeypatch.setattr(argparse._sys, "argv", [self.argv0])

        with pytest.raises(SystemExit) as exc_info:
            parse_args()

        assert str(exc_info.value) == "1"

        out, err = capsys.readouterr()
        assert "usage: ttfautohint" in out
        assert not err

    def test_no_out_file(self, monkeypatch, capsys):
        monkeypatch.setattr(argparse._sys, "argv", [self.argv0, "font.ttf"])

        with pytest.raises(SystemExit) as exc_info:
            parse_args()

        assert str(exc_info.value) == "1"

        out, err = capsys.readouterr()
        assert "usage: ttfautohint" in out
        assert not err

    def test_source_date_epoch(self, monkeypatch):
        epoch = "1513966552"
        env = dict(os.environ)
        env["SOURCE_DATE_EPOCH"] = epoch
        monkeypatch.setattr(os, "environ", env)

        options = parse_args([])

        assert options["epoch"] == int(epoch)

    def test_source_date_epoch_invalid(self, monkeypatch):
        invalid_epoch = "foobar"
        env = dict(os.environ)
        env["SOURCE_DATE_EPOCH"] = invalid_epoch
        monkeypatch.setattr(os, "environ", env)

        with pytest.warns(UserWarning, match="invalid SOURCE_DATE_EPOCH: 'foobar'"):
            options = parse_args([])

        assert "epoch" not in options

    def test_show_ttfa_info_unsupported(self):
        with pytest.raises(NotImplementedError):
            parse_args("-T")

    def test_parse_args_custom_splitfunc(self):
        # https://github.com/fonttools/ttfautohint-py/issues/2
        s = '"build folder\\unhinted\\Test_Lt.ttf" "output folder\\hinted\\Test_Lt.ttf"'
        args = parse_args(s, splitfunc=_windows_cmdline2list)
        assert args["in_file"] == "build folder\\unhinted\\Test_Lt.ttf"
        assert args["out_file"] == "output folder\\hinted\\Test_Lt.ttf"

    @pytest.mark.skipif(sys.platform != "win32", reason="only for windows")
    def test_parse_args_windows_paths(self):
        # https://github.com/fonttools/ttfautohint-py/issues/2
        s = '"build folder\\unhinted\\Test_Lt.ttf" "output folder\\hinted\\Test_Lt.ttf"'
        args = parse_args(s)
        assert args["in_file"] == "build folder\\unhinted\\Test_Lt.ttf"
        assert args["out_file"] == "output folder\\hinted\\Test_Lt.ttf"



class TAError(Exception):
    def __init__(self, rv, error_string):
        self.rv = int(rv)
        self.error_string = error_string.decode("utf-8", errors="replace")

    def __str__(self):
        error = self.rv
        error_string = self.error_string
        s = "0x%02X" % error
        if error_string:
            s += ": %s" % error_string
        return s



import sys
import ttfautohint


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return ttfautohint.run(args).returncode


# The following constants are only kept only for backward-compatibility
# to support the deprecated ttfautohint.options.parse_args function
USAGE = "ttfautohint [OPTION]... [IN-FILE [OUT-FILE]]"

DESCRIPTION = """\
Replace hints in TrueType font IN-FILE and write output to OUT-FILE.
If OUT-FILE is missing, standard output is used instead;
if IN-FILE is missing also, standard input and output are used.

The new hints are based on FreeType's auto-hinter.

This program is a simple front-end to the `ttfautohint' library.
"""

EPILOG = """\
The program accepts both TTF and TTC files as input.
Use option -i only if you have a legal permission to modify the font.
The used PPEM value for option -p is FUnits per em, normally 2048.
With option -s, use default values for standard stem width and height,
otherwise they are derived from script-specific characters
resembling the shape of character `o'.

A hint set contains the optimal hinting for a certain PPEM value;
the larger the hint set range (as given by options -l and -r),
the more hint sets get computed, usually increasing the output font size.
The `gasp' table of the output file always enables grayscale hinting
for all sizes (limited by option -G, which is handled in the bytecode).
Increasing the value of -G does not increase the output font size.

Options -f and -D take a four-letter string that identifies a script.
Option -f sets the script used as a fallback for glyphs that can't be
associated with a known script.  By default, such glyphs are hinted;
if option -S is set, they are scaled only instead.  Option -D sets the
default script for handling OpenType features.

Possible four-letter string values are

  adlm (Adlam),
  arab (Arabic),
  armn (Armenian),
  avst (Avestan),
  bamu (Bamum),
  beng (Bengali),
  buhd (Buhid),
  cakm (Chakma),
  cans (Canadian Syllabics),
  cari (Carian),
  cher (Cherokee),
  copt (Coptic),
  cprt (Cypriot),
  cyrl (Cyrillic),
  deva (Devanagari),
  dsrt (Deseret),
  ethi (Ethiopic),
  geor (Georgian (Mkhedruli)),
  geok (Georgian (Khutsuri)),
  glag (Glagolitic),
  goth (Gothic),
  grek (Greek),
  gujr (Gujarati),
  guru (Gurmukhi),
  hebr (Hebrew),
  kali (Kayah Li),
  khmr (Khmer),
  khms (Khmer Symbols),
  knda (Kannada),
  lao (Lao),
  latn (Latin),
  latb (Latin Subscript Fallback),
  latp (Latin Superscript Fallback),
  lisu (Lisu),
  mlym (Malayalam),
  mong (Mongolian),
  mymr (Myanmar),
  nkoo (N'Ko),
  olck (Ol Chiki),
  orkh (Old Turkic),
  osge (Osage),
  osma (Osmanya),
  saur (Saurashtra),
  shaw (Shavian),
  sinh (Sinhala),
  sund (Sundanese),
  taml (Tamil),
  tavt (Tai Viet),
  telu (Telugu),
  tfng (Tifinagh),
  thai (Thai),
  vaii (Vai),
  none (no script).

A control instructions file contains entries of the form

  [<font idx>] <script> <feature> @ <glyph ids>

  [<font idx>] <script> <feature> w <stem widths>

  [<font idx>] <glyph id> l|r <points> [(<left offset>,<right offset>)]

  [<font idx>] <glyph id> n <points>

  [<font idx>] <glyph id> t|p <points> [x <shift>] [y <shift>] @ <ppems>

<font idx> is the current subfont, <glyph id> is a glyph name or index,
<glyph ids> is a set of <glyph id>s, <stem widths> is an unordered set of
integer stem widths in font units, <shift> is a real number in px,
<points> and <ppems> are integer ranges as with option `-X'.

<script> and <feature> are four-letter tags that define a style
the <glyph ids> are assigned to; possible values for <script> are the same
as with option -D, possible values for <feature> are

  c2cp (petite capitals from capitals),
  c2sc (small capitals from capitals),
  ordn (ordinals),
  pcap (petite capitals),
  ruby (ruby),
  sinf (scientific inferiors),
  smcp (small capitals),
  subs (subscript),
  sups (superscript),
  titl (titling).

`w' assigns stem widths to a style; the first value sets the default.
`l' (`r') creates one-point segments with direction left (right).
<left offset> and <right offset> specify offsets (in font units)
relative to the corresponding points to give the segments a length.
`n' removes points from horizontal segments, making them `weak' points.
`t' (`p') applies delta exceptions to the given points before (after) IUP.

`#' starts a line comment, which gets ignored.
Empty lines are ignored, too.

Key letters `l', `r', `n', `p', `t', `w', `x', and `y'
have the verbose aliases `left', `right', `nodir', `point', `touch',
`width', `xshift', and `yshift', respectively.

Report bugs to: freetype-devel@nongnu.org

ttfautohint home page: <http://www.freetype.org/ttfautohint>
"""



from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.bdist_wheel import bdist_wheel
from distutils.command.clean import clean
from distutils.file_util import copy_file
from distutils.dir_util import mkpath
from distutils import log
import os
import platform
import subprocess


class UniversalBdistWheel(bdist_wheel):
    def get_tag(self):
        return ("py3", "none") + bdist_wheel.get_tag(self)[2:]


class Executable(Extension):
    if os.name == "nt":
        suffix = ".exe"
    else:
        suffix = ""

    def __init__(self, name, output_dir=".", cwd=None, env=None):
        Extension.__init__(self, name, sources=[])
        self.target = self.name.split(".")[-1] + self.suffix
        self.output_dir = output_dir
        self.cwd = cwd
        self.env = env


class ExecutableBuildExt(build_ext):
    def get_ext_filename(self, ext_name):
        for ext in self.extensions:
            if isinstance(ext, Executable):
                return os.path.join(*ext_name.split(".")) + ext.suffix
        return build_ext.get_ext_filename(self, ext_name)

    def build_extension(self, ext):
        if not isinstance(ext, Executable):
            build_ext.build_extension(self, ext)
            return

        if platform.system() == "Windows":
            # we need to run make from a bash shell.
            cmd = ["bash", "-c", "make all"]
        else:
            cmd = ["make", "all"]

        log.debug("running '{}'".format(" ".join(cmd)))
        if not self.dry_run:
            env = dict(os.environ)
            if ext.env:
                env.update(ext.env)
            if platform.system() == "Windows":
                import struct

                # MSYS2 is required on Windows, as we need bash, make, autotools etc.
                msys2_root = os.path.abspath(env.get("MSYS2ROOT", "C:\\msys64"))
                msys2_bin = os.path.join(msys2_root, "usr", "bin")
                if not os.path.isdir(msys2_bin):
                    from distutils.errors import DistutilsPlatformError

                    raise DistutilsPlatformError(f"Could not find {msys2_bin}")

                # select mingw32 or mingw64 toolchain depending on python architecture
                bits = struct.calcsize("P") * 8
                toolchain = "mingw%d" % bits
                msys2_mingw_bin = os.path.join(msys2_root, toolchain, "bin")
                extra_paths = [msys2_mingw_bin, msys2_bin]

                # See if we can use the standalone MinGW with win32 threads instead of MSYS2's
                # (the latter only comes with posix threads and unnecessarily pulls in
                # libwinpthread-1.dll)
                mingw_root = env.get("MINGWROOT")
                if mingw_root is not None:
                    mingw_root = os.path.abspath(mingw_root)
                    standalone_mingw = os.path.join(mingw_root, toolchain, "bin")
                    if not os.path.isdir(standalone_mingw):
                        from distutils.errors import DistutilsPlatformError

                        raise DistutilsPlatformError(
                            f"Could not find {standalone_mingw}"
                        )
                    extra_paths.insert(0, standalone_mingw)

                PATH = os.pathsep.join([*extra_paths, env["PATH"]])
                env.update(
                    PATH=PATH,
                    MSYSTEM=toolchain.upper(),
                    # this tells bash to keep the current working directory
                    CHERE_INVOKING="1",
                )

            if self.force:
                subprocess.call(["make", "clean"], cwd=ext.cwd, env=env)
            p = subprocess.run(cmd, cwd=ext.cwd, env=env, shell=True)
            if p.returncode != 0:
                from distutils.errors import DistutilsExecError

                raise DistutilsExecError("running 'make' failed")

        exe_fullpath = os.path.join(ext.output_dir, ext.target)

        dest_path = self.get_ext_fullpath(ext.name)
        mkpath(os.path.dirname(dest_path), verbose=self.verbose, dry_run=self.dry_run)

        copy_file(exe_fullpath, dest_path, verbose=self.verbose, dry_run=self.dry_run)


class CustomClean(clean):
    def run(self):
        clean.run(self)
        if not self.dry_run:
            # if -a, also reset submodules to remove all the build byproducts
            if self.all:
                subprocess.run(
                    [
                        "git",
                        "submodule",
                        "foreach",
                        "--recursive",
                        "git",
                        "clean",
                        "-fdx",
                    ],
                    check=True,
                )
                subprocess.run(
                    [
                        "git",
                        "submodule",
                        "foreach",
                        "--recursive",
                        "git",
                        "reset",
                        "--hard",
                    ]
                )
                subprocess.run(
                    [
                        "git",
                        "submodule",
                        "update",
                        "--init",
                        "--recursive",
                    ],
                    check=True,
                )
            subprocess.run(["make", "clean"], cwd=os.path.join("src", "c"), check=True)


ttfautohint_exe = Executable(
    "ttfautohint.ttfautohint",
    cwd=os.path.join("src", "c"),
    output_dir=os.path.join("build", "local", "bin"),
)

cmdclass = {}
ext_modules = []
for env_var in ("TTFAUTOHINTPY_BUNDLE_DLL", "TTFAUTOHINTPY_BUNDLE_EXE"):
    if os.environ.get(env_var, "0") in {"1", "yes", "true"}:
        cmdclass["bdist_wheel"] = UniversalBdistWheel
        cmdclass["build_ext"] = ExecutableBuildExt
        cmdclass["clean"] = CustomClean
        ext_modules = [ttfautohint_exe]

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name="ttfautohint-py",
    use_scm_version={"write_to": "src/python/ttfautohint/_version.py"},
    description=(
        "Python wrapper for ttfautohint, a free auto-hinter for TrueType fonts"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Cosimo Lupo",
    author_email="cosimo@anthrotype.com",
    url="https://github.com/fonttools/ttfautohint-py",
    license="MIT",
    platforms=["posix", "nt"],
    package_dir={"": "src/python"},
    packages=find_packages("src/python"),
    ext_modules=ext_modules,
    zip_safe=True,
    cmdclass=cmdclass,
    setup_requires=["setuptools_scm"],
    extras_require={"testing": ["pytest", "coverage", "fontTools"]},
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Text Processing :: Fonts",
        "Topic :: Multimedia :: Graphics",
    ],
)



def ensure_binary(s, encoding="ascii", errors="strict"):
    if isinstance(s, str):
        return s.encode(encoding, errors)
    elif isinstance(s, bytes):
        return s
    else:
        raise TypeError("not expecting type '%s'" % type(s))


def ensure_text(s, encoding="ascii", errors="strict"):
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    elif isinstance(s, str):
        return s
    else:
        raise TypeError("not expecting type '%s'" % type(s))
